/*
 * UNIQUE FILE FOR PVM, MPI, TRANSPUTER AND SEQUENTIAL COMPILING
 * FILE: llcsync.h
 * LAST REV: 21/8/1998
 *
 * NOTES:
 *      - (MPI) The PAR function uses MPI_Sendrecv to exchange data. Thus, there
 *        __cannot__ be aliasing between the send and receive buffers.
 * 	- This file must be included with at least CCMPI,
 * 	  CCPVM or CCTRANSPUTER defined (depending on topology).
 * 	  There is also a compilation for a sequential model (CCSEQ), for
 * 	  debugging purposes. It just defines everything in blank.
 * 	- Note that certain function have a note: "G function". They are functions with
 * 	  intended to return the same value for every processor in the VM (such as
 * 	  GRAND). All such functions start with a 'G'.
 * 	- Any try to print more than 1024 chars will produce a core dump (the
 * 	  temporal buffer length exceeded)
 * 	- Sync issues make impossible to put a LLC call for only one processor.
 * 	  Things like "if (LLNAME == 2) GPRINTF..." would hang the transputer.
 * 	  However, there are special LLC functions for these purposes, such as
 * 	  GSEND and BPRINTF.
 *      - (MPI) There is a buffering system implemented. To use it, #define
 *        LL_SIZEBUFFER with something greater than 0. This value (if set)
 *        can be modified in the command line with the parameter -mem <BUFSIZE>
 *      - (PVM) You can modify the number of processors in the PVM version
 *        with the parameter -np <numprocessors>, but numprocessors will have
 *        to take a value less than (1 << DIM), specified at compile time.
 *
 * LAST CHANGES:
 *      (9-aug-98) Integrated MPI version
 *      (9-aug-98) numt and my_parent made local to PVM INITIALIZE
 *      (17-aug-98) old PARx macros fixed
 *      (17-aug-98) pvm_initsend removed from SEND macros
 *      (18-aug-98) GGETx routines fixed for working with MPI
 *      (19-aug-98) Solved problems with Digital ARCH (i.e.: # not in column 1).
 *      (21-aug-98) INITIALIZE_DIM repaired for PVM
 */

/*
#define CCMPI
*/
/**/

#ifndef LLCINCLUDED
#define LLCINCLUDED

#if !defined(CCPVM) && !defined(CCTRANSPUTER) && !defined(CCSEQ) && !defined(CCMPI)
#   error Compilation topology not defined (CCPVM, CCTRANSPUTER, CCSEQ, CCMPI)
#endif

/******** Include zone ********/

/* --- Standard includes --- */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <stdarg.h>

/* --- Conditional includes --- */
#ifdef CCTRANSPUTER
#   include <process.h>
#   include <channel.h>
#   include <misc.h> /* get_param */
#endif

#ifdef CCPVM
#   include "pvm3.h"
#endif

#ifdef CCMPI
#   include "mpi.h"
int DIM;
#else
#   include "dimensio.h"
#endif
/******* End of include zone ********/


/******* Definition zone ************/
/* --- Common defs --- */
#ifndef TRUE
#   define TRUE     (1==1)
#   define FALSE    (0==1)
#endif
#define BOOLEAN  int

#ifndef LLSIZE
#   define LLSIZE 1        /* Size of data for macros such as the PAR func */
#endif

#ifndef SIZEBUFFER
#   define SIZEBUFFER 1024 /* For a standard array */
#endif

/* --- Sequential defs and vars --- */
#ifdef CCSEQ
	
#undef DIM
#   define DIM 1
#   define INITIALIZE	
#   define EXIT return(0)
#   define LLNUMPROCESSORS 1
#   define PRINTF printf
#   define GPRINTF printf
#   define PPRINTF
#   define BPRINTF
#   define PAR(f1, p1, s1, f2, p2, s2) f1; f2;
#   define NAME 1
#   define GROUPNAME 0	
#   define GRAND(x) x = rand()
#   define SENDCHAR(lldim, llint)
#   define SENDINT(lldim, llint)
#   define SENDFLOAT(lldim, llfloat)
#   define SEND(lldim, lldata, llsize)
#   define RECEIVECHAR(lldim, llchar)
#   define RECEIVEINT(lldim, llint)
#   define RECEIVEFLOAT(lldim, llfloat)
#   define RECEIVE(lldim, lldata, llsize)

	int LLNAME = 0;
	int bit = DIM;
	int NUMPROCESSORS = LLNUMPROCESSORS;
#else  /* --- Until the end of the file --- */


#ifdef ROOT
#   define ISROOT 1
#else
#   define ISROOT 0 /* Is my processor the ROOT processor? */
#endif

#define INROOTSET (LLNAME < NUMPROCESSORS)  /* Am I in the ROOT set? */
#define IAMLAST ((LLNAME & (NUMPROCESSORS - 1)) == (NUMPROCESSORS - 1)) /* Am I the last processor in my group? */
#define GROUPNAME (NAME >> bit)
#define LLGROUPNAME (LLNAME >> bit)         /* The name of the group I belong to */
#define LLPARTNER(i) (LLNAME ^ (1 << i))    /* My partner on the other side */
#define NAME (LLNAME & (NUMPROCESSORS - 1)) /* My name in my actual group */
#define LLNUMPROCESSORS (1 << DIM)          /* # of processors in the whole machine */

typedef struct {
	int procdest;     /* Destination of the message */
	int sizemessage;  /* Size of the message */
	} LLROUTE;        /* - Header of a message between processors in a GSEND - */

/* --- Common defs and vars --- */
  int NUMPROCESSORS;      /* # of Processors in the group */
  int LLNAME;                               /* Processor identifier */
  int bit;                                  /* Actual dimension */
  int level;                                /* Processor level in the Hypercube */

/* --- PVM defs and vars --- */
#ifdef CCPVM

  /* - Types of Message Passing - */
#   define ENCODING PvmDataRaw
#   define MSG_TIDS      100
#   define MSG_ASSIGN    101
#   define MSG_REDU      102
#   define MSG_BROADCAST 103
#   define MSG_END	 104

  int mytid;                     /* my task id */
  int tids[LLNUMPROCESSORS];     /* slave task ids */
  int atid, atag, alen;          /* to be used with pvm_precv  */
  /*int *ti = tids;  I think we don't use this, anyway (NACHO) */
  int my_parent;
  void INITIALIZE_DIM(int dim);
#endif

/* --- TRANSPUTER defs and vars --- */
#ifdef CCTRANSPUTER

  int dim;                       /* current dimension */
  Channel **d_in = NULL;         /* Input Communication Channels  */
  Channel **d_out = NULL;        /* Output Communication Channels */

#endif

/* --- MPI defs and vars --- */
#ifdef CCMPI

/*
 * #define LL_SIZEBUFFER with the size of the buffer for MPI Send & Receive,
 * or 0 to use MPI internal buffer.
 * Note that, with LL_SIZEBUFFER > 0, its size can be modified in the
 * command line with a -mem <BUFSIZE> parameter.
 */
#ifndef LL_SIZEBUFFER
#  define LL_SIZEBUFFER  2000  /* Size of MPI buffer */
#endif

#if (LL_SIZEBUFFER == 0)       /* Don't do buffering */
#  define LLSEND MPI_Send
#  define LL_REGISTER_BUF
#  define LL_UNREGISTER_BUF
#  define ll_buffer_size LL_SIZEBUFFER 
#else                          /* Yes Do buffering */
#  define LLSEND MPI_Bsend
int ll_buffer_size;            /* Buffer size */
char *ll_buffer;               /* Buffer */
#  define LL_REGISTER_BUF \
  ll_buffer_size = setbufsize(argc, argv); \
  if ((ll_buffer = malloc(ll_buffer_size)) == NULL) { \
     if (LLNAME == 0)                                       \
        printf("Error: Not enough memory for message buffers. Exiting.\n");\
     MPI_Finalize();        \
     return 1;              \
  }                         \
  MPI_Buffer_attach(ll_buffer, ll_buffer_size)

#  define LL_UNREGISTER_BUF \
      MPI_Buffer_detach(ll_buffer, &ll_buffer_size); \
      free(ll_buffer)

#endif /* if (LL_SIZEBUFFER == 0) */

MPI_Status status;             /* Status for each MPI_Recv */
MPI_Comm LLC_COMM;             /* LLC communicator */
#define LLCTAG 0               /* Tag for LLC communications */

#endif /* #ifdef CCMPI */

/******* End of types, vars and defs zone ******/

/******** Level 0 functions ********/

#if defined(CCPVM) || defined(CCMPI)
/*
 *  Calculates log2 of a number
 */
int llclog2(int n)
{
int i;
 for (i = 16; i > 0; i--)
   if (n & (1 << i))
     return i;
return 0;
}
#endif

/* --- PVM Level 0 functions --- */
#ifdef CCPVM
#  define INITIALIZE(argc, argv) \
   { \
   int n, numproc; \
			PRINTF(" Starting ...\n"); \
      for (n = 1; (n < argc); n++) \
         if (!strncmp("-np", argv[n], 3)) break; \
      if (n == argc) \
	     INITIALIZE_DIM(DIM);  \
      else { \
	     numproc = atoi(argv[n]+3); \
        if (!numproc && (n < argc - 1)) \
	        numproc = atoi(argv[n+1]); \
        if (numproc < 2) { \
	       PRINTF("Error: specified \"-np\" without a valid number of processors\n"); \
	       exit(1); \
	     } \
        if (numproc > (1 << DIM)) { \
	        PRINTF("Warning: number of processors greater than DIM. %d assumed\n", 1 << DIM); \
	        numproc = 1 << DIM; \
	     } \
        NUMPROCESSORS = numproc; \
        INITIALIZE_DIM(llclog2(numproc)); \
      } \
   }

#  define EXIT GPRINTF(""); pvm_exit(); exit(0)
#  define SENDCHAR(lldim, x) \
	 pvm_psend(tids[LLPARTNER(lldim)], LLNAME, &x, 1, PVM_BYTE)
#  define SENDINT(lldim, llint) \
	 pvm_psend(tids[LLPARTNER(lldim)], LLNAME, &llint, 1, PVM_INT)
#  define SENDFLOAT(lldim, llfloat)  \
	 pvm_psend(tids[LLPARTNER(lldim)], LLNAME, &llfloat, 1, PVM_FLOAT)
#  define SEND(lldim, lldata, llsize) \
	 pvm_psend(tids[LLPARTNER(lldim)], LLNAME, (char *)(lldata), llsize, PVM_BYTE)
#  define SENDFILE(lldim, lldata, llsize) \
	 pvm_psend(tids[LLPARTNER(lldim)], LLNAME, (FILE *)(lldata), llsize, PVM_BYTE)
#  define RECEIVEFILE(lldim, lldata, llsize)  \
	 pvm_precv(tids[LLPARTNER(lldim)], LLPARTNER(lldim), (FILE *)(lldata), llsize, PVM_BYTE, &atid, &atag, &alen)
#  define RECEIVECHAR(lldim, llchar) \
	 pvm_precv(tids[LLPARTNER(lldim)], LLPARTNER(lldim), &llchar, 1, PVM_BYTE, &atid, &atag, &alen)
#  define RECEIVEINT(lldim, llint) \
	 pvm_precv(tids[LLPARTNER(lldim)], LLPARTNER(lldim), &llint, 1, PVM_INT, &atid, &atag, &alen)
#  define RECEIVE(lldim, lldata, llint) \
	 pvm_precv(tids[LLPARTNER(lldim)], LLPARTNER(lldim), (char *)(lldata), llint, PVM_BYTE, &atid, &atag, &alen)
#  define RECEIVEFLOAT(lldim, llfloat) \
	 pvm_precv(tids[LLPARTNER(lldim)], LLPARTNER(lldim), &llfloat, 1, PVM_FLOAT, &atid, &atag, &alen)
#  define GSEND(org, dest, lldata, llsize) { \
	 if (LLNAME == org) { \
		 pvm_psend(tids[dest], LLNAME, &llsize, 1, PVM_INT); \
		 pvm_psend(tids[dest], LLNAME, (char *)(lldata), llsize, PVM_BYTE); \
		 } \
	 if (LLNAME == dest) { \
		 pvm_precv(tids[org], org, &llsize, 1, PVM_INT, &atid, &atag, &alen); \
		 pvm_precv(tids[org], org, &lldata, 1, llsize, &atid, &atag, &alen); \
		 } \
	 }
#endif

/* --- TRANSPUTER Level 0 functions --- */
#ifdef CCTRANSPUTER
#  define EXIT GPRINTF (""); exit_terminate (0)
#  define SENDCHAR(lldim, llint) ChanOutChar (d_out[lldim], llint)
#  define SENDINT(lldim, llint) ChanOutInt (d_out[lldim], llint)
#  define SENDFLOAT(lldim, llfloat) ChanOut (d_out[lldim], &llfloat, sizeof(llfloat))
#  define SEND(lldim, lldata, llsize) ChanOut (d_out[lldim], lldata, llsize)
#  define RECEIVECHAR(lldim, llchar) llchar = ChanInChar (d_in[lldim])
#  define RECEIVEINT(lldim, llint) llint = ChanInInt (d_in[lldim])
#  define RECEIVEFLOAT(lldim, llfloat) ChanIn(d_in[lldim], &llfloat, sizeof(llfloat))
#  define RECEIVE(lldim, lldata, llsize) ChanIn (d_in[lldim], lldata, llsize)
#  define GSEND(org, dest, lldata, llsize) if (org != dest) { \
		int llnode, lli, llcurbit;  \
		for (llnode = org, lli = 0; lli < bit; lli++) {  \
		  llcurbit = 1 << lli;  \
		  if ((llnode & llcurbit) ^ (dest & llcurbit)) {  \
				  if (LLNAME == llnode) { \
					  SENDINT(lli, llsize); \
					  SEND(lli, lldata, llsize);  \
					  } \
				  if (LLNAME == (llnode ^ llcurbit)) { \
					  RECEIVEINT(lli, llsize); \
					  RECEIVE(lli, lldata, llsize);  \
					  } \
				  llnode ^= llcurbit;  \
				  } \
		  }  \
		}

#endif

/* --- MPI Level 0 functions --- */
#ifdef CCMPI
/* --- Return the size for the message buffer ---
 * RECEIVES: The argc and argv params of the main func.
 * RETURN: If there is not a -mem parameter, it returns LL_SIZEBUFFER.
 *         If there is, returns the -mem value.
 */
int setbufsize(int argc, char **argv)
{
int n, sizebuffer;

  for (n = 1; n < argc; n++) \
     if (!strncmp("-mem", argv[n], 4)) break; \
  if (n == argc)
     return LL_SIZEBUFFER;
  sizebuffer = atoi(argv[n]+4);
  if (!sizebuffer && (n < argc - 1))
     sizebuffer = atoi(argv[n+1]);
  if (sizebuffer < 100) {
    printf("Warning: Buffer size < 100 ignored. Taken %i\n", LL_SIZEBUFFER);
    return LL_SIZEBUFFER;
  }
  return sizebuffer;
}

/* --- Initializes the LLC communicator ---
 * RECEIVES: The number of processors in the mpirun call
 * RETURNS: Initializes the LLC_COMM
 * NOTES: The returned communicator will only work for NUMPROCESSORS
 * processors. If you try to execute, say, "mpirun -np 7 my_program", the
 * hypercube will only take 4 processors, NUMPROCESSORS will be set to 4
 * and any communication will only take care of 4 processors. The rest will
 * take undefined behaviour.
 * Note that this procedure works with a maximum of 64 processors. For
 * bigger architectures, increase the array size.
 */
void initcomm(int numprocessors)
{
int tids[64], n;
MPI_Group world,           /* Group identifier for MPI_COMM_WORLD */
          hypercube;       /* Group identifier for my hypercube   */

  for (n = 0; n < numprocessors; n++)
    tids[n] = n;
  MPI_Comm_group(MPI_COMM_WORLD, &world);
  MPI_Group_incl(world, numprocessors, tids, &hypercube);
  MPI_Comm_create(MPI_COMM_WORLD, hypercube, &LLC_COMM);
}

#define INITIALIZE(argc, argv) { \
  MPI_Init( &argc, &argv ); \
  MPI_Comm_rank(MPI_COMM_WORLD, &LLNAME); \
  MPI_Comm_size(MPI_COMM_WORLD, &NUMPROCESSORS); \
  LL_REGISTER_BUF; \
  DIM = llclog2(NUMPROCESSORS); \
  bit = DIM; \
  if (NUMPROCESSORS != (1 << bit)) { \
     if (LLNAME == 0) \
        printf("Warning: specified %i processors, but the hipercube takes only %i.\n", NUMPROCESSORS, 1 << bit); \
     NUMPROCESSORS = 1 << bit; \
     } \
  initcomm(NUMPROCESSORS); \
  if (LLNAME == 0) \
     printf("LLC initialized. MPI version, buffer size: %d %s", ll_buffer_size,\
            (LL_SIZEBUFFER == 0)? "(deactivated)\n" : "\n"); \
}

#define EXIT { \
  MPI_Finalize(); \
  LL_UNREGISTER_BUF; \
  }

#define SENDCHAR(lldim, llchar) \
  LLSEND(&llchar, 1, MPI_CHAR, LLPARTNER(lldim), LLCTAG, LLC_COMM)
#define SENDINT(lldim, llint) \
  LLSEND(&llint, 1, MPI_INT, LLPARTNER(lldim), LLCTAG, LLC_COMM)
#define SENDFLOAT(lldim, llfloat) \
  LLSEND(&llfloat, 1, MPI_FLOAT, LLPARTNER(lldim), LLCTAG, LLC_COMM)
#define SEND(lldim, lldata, llsize) \
  LLSEND(lldata, llsize, MPI_BYTE, LLPARTNER(lldim), LLCTAG, LLC_COMM)
#define RECEIVECHAR(lldim, llchar) \
  MPI_Recv(&llchar, 1, MPI_CHAR, LLPARTNER(lldim), LLCTAG, LLC_COMM, &status)
#define RECEIVEINT(lldim, llint) \
  MPI_Recv(&llint, 1, MPI_INT, LLPARTNER(lldim), LLCTAG, LLC_COMM, &status)
#define RECEIVEFLOAT(lldim, llfloat) \
  MPI_Recv(&llfloat, 1, MPI_FLOAT, LLPARTNER(lldim), LLCTAG, LLC_COMM, &status)
#define RECEIVE(lldim, lldata, llsize) \
  MPI_Recv(lldata, llsize, MPI_BYTE, LLPARTNER(lldim), LLCTAG, LLC_COMM, &status)
#define GSEND(org, dest, lldata, llsize) { \
	 if (LLNAME == org)  \
       LLSEND(lldata, llsize, MPI_BYTE, dest, LLCTAG, LLC_COMM) \
	 if (LLNAME == dest) \
       MPI_Recv(lldata, llsize, MPI_BYTE, org, LLCTAG, LLC_COMM, &status) \
}

#endif

/******** End of Level 0 functions ********/


/******** Level 1 functions ********/

/* --- Definitions about the ROOT process (PVM, TRANSPUTER) --- */
/* NOTE: In MPI, ROOT shall _NOT_ be defined */
#ifdef ROOT

	/* INITIALIZE: Initialize the Parallel machine
	 *	Receives: Nothing
	 * Returns: Initializes the whole stuff. Must be the first LLC call
	 */
	/* --- PVM INITIALIZE --- */
#        ifdef CCPVM
        /* --- Initializes the PVM Virtual Machine ---
         * RECEIVES: The dimension of the VM
	 * NOTES: Not to be called directly, but from the INITIALIZE & INITPROCESSORS macros
	 */
         void INITIALIZE_DIM(int dim) {
	 int ll_inum, ll_i;
	 int numt;
	    mytid = pvm_mytid();
	    my_parent = -1;
            NUMPROCESSORS = 1 << dim;
	    printf("Initializing. NUMPROCESSORS= %d\n", NUMPROCESSORS);
	    LLNAME = 0;
	    level = node_level (LLNAME);
	    bit = dim;  
	    tids[0] = mytid; 
	    numt = pvm_spawn("node", (char**)0, 0, "", NUMPROCESSORS - 1, (tids + 1)); 
	    if (numt < (NUMPROCESSORS - 1)) 
	      { 
		printf("Trouble spawning slaves. Aborting. Num %d, Error codes are:\n", numt);
		for (ll_i = numt; ll_i < (NUMPROCESSORS - 1); ll_i++ )
		  {
		    printf("TID %d %d\n", ll_i, tids[ll_i]);
		  }
		for (ll_i = 0; ll_i < (NUMPROCESSORS - 1); ll_i++ )
		  {
		    pvm_kill(tids[ll_i]);
		  }
		pvm_exit();
		exit(-1);
	      }
	    pvm_initsend(ENCODING);
	    pvm_pkint(tids, NUMPROCESSORS, 1);
	    pvm_mcast(&tids[1], NUMPROCESSORS-1, MSG_TIDS);
	 }

#        endif /* PVM INITIALIZE */

	/* --- TRANSPUTER INITIALIZE --- */
#        ifdef CCTRANSPUTER
#               define INITIALIZE \
		{ \
			set_abort_action (ABORT_HALT); \
			dim = *((int*) get_param (5)); \
			d_in = CreateChannels (get_param (3), dim); \
			d_out = CreateChannels (get_param (4), dim); \
			LLNAME = *((int*) get_param (6)); \
			level = node_level (LLNAME); \
			bit = dim; \
			NUMPROCESSORS = (1 << dim); \
		}
#        endif /* TRANSPUTER INITIALIZE */
/* --- End of defs about the root process (PVM, TRANSPUTER) --- */

#else
/* --- Definitions about the child processes ---
 * The meaning and parameters of each function is on the ROOT definition of
 *	the same function
 */

	/* --- PVM INITIALIZE --- */
#  ifdef CCPVM
   void INITIALIZE_DIM(int dim) {
       int ll_inum, ll_i;
       NUMPROCESSORS = 1 << dim;
       mytid = pvm_mytid();
       my_parent = pvm_parent();
       pvm_recv(my_parent, MSG_TIDS);
       pvm_upkint(tids, NUMPROCESSORS, 1);
       bit = dim;
	    for(ll_i = 0; ll_i < NUMPROCESSORS; ll_i++)
	      if (mytid == tids[ll_i])
            {
              LLNAME = ll_i;
              level = node_level (LLNAME);
              break;
            }
   }
#      endif /* PVM INITIALIZE */

     /* --- TRANSPUTER INITIALIZE --- */
#        ifdef CCTRANSPUTER
#               define INITIALIZE \
		{ \
			set_abort_action (ABORT_HALT); \
			dim = *((int*) get_param (3)); \
			d_in = CreateChannels (get_param (1), dim); \
			d_out = CreateChannels (get_param (2), dim); \
			LLNAME = *((int*) get_param (4)); \
			level = node_level (LLNAME); \
			bit = dim; \
			NUMPROCESSORS = (1 << dim); \
		}
#        endif /* TRANSPUTER INITIALIZE */
#endif
/* --- End of definitions about the child processes --- */

/* --- Gets a char from the stdin ---
 * RECEIVES: Where to put the char f
 * RETURNS: The char f
 */
#define GETC(f) \
{ \
    int lli; \
    if (LLNAME == 0) { \
       f = getc (stdin); \
       lli = getc (stdin); \
    } \
    else \
       f = 0; \
}

/* --- Calculates the same random number for all the processors ---
 * RECEIVES: Where to put the random number, x
 * RETURNS: the random number, x
 * G function
 */
#define GRAND(x) \
{ \
  int llj; \
  if (LLNAME == 0) { \
     x = rand (); \
     for (llj = 0; llj < bit; llj++) \
         SENDINT(llj, x); \
   } \
   else { \
      RECEIVEINT(level-1, x); \
      for (llj=level;llj<bit;llj++) \
	  SENDINT(llj, x); \
   } \
}

/* --- Gets a char/int/float/string from the stdin ---
 * RECEIVES: Where to put the char f
 * RETURNS: The char f
 * G function
 */
#define GGETC(x) \
{ \
  int lli, llaux; \
  if (LLNAME == 0) { \
     x = getc (stdin); \
     llaux = getc (stdin); \
     for (lli = 0; lli < bit; lli++) \
	 SENDCHAR(lli, x); \
   } \
   else { \
     RECEIVECHAR(level - 1, x); \
     for (lli = level; lli < bit; lli++) \
	 SENDCHAR(lli, x); \
   } \
}

#define GGETI(x) \
{ \
  int lli, llaux; \
  if (LLNAME == 0) { \
     scanf ("%d", &x); \
     llaux = getc (stdin); \
     for (lli = 0; lli < bit; lli++) \
	 SENDINT(lli, x); \
  } \
  else { \
     RECEIVEINT(level-1, x); \
     for (lli = level; lli < bit; lli++) \
	 SENDINT(lli, x); \
  } \
}

#define GGETF(x) \
{ \
  int lli, llaux; \
  if (LLNAME == 0) { \
     scanf ("%f", &x); \
     llaux = getc (stdin); \
     for (lli = 0; lli < bit; lli++) \
	 SENDFLOAT(lli, x); \
  } \
  else { \
     RECEIVEFLOAT(level-1, x); \
     for (lli = level; lli < bit; lli++) \
	 SENDFLOAT(lli, x); \
  } \
}

#define GGETS(x) \
{ \
  int lli, llaux, lllen; \
  if (LLNAME == 0) { \
     gets (x); \
     llaux = strlen(x)+1; \
     for (lli = 0; lli < bit; lli++) \
     { \
       SENDINT(lli, llaux); \
       SEND(lli, x, llaux); \
     } \
  } \
  else { \
     RECEIVEINT(level-1, lllen); \
     RECEIVE(level-1, x, lllen); \
     for (lli = level; lli < bit; lli++) \
     { \
	SENDINT(lli, lllen); \
	SEND(lli, x, lllen); \
     } \
   } \
}


/******* End of Level 1 Functions ******/


/******* Level 2 Functions (parallel data interchange related) ******/

/* --- Paralell Fork ---
 * RECEIVES: Two func calls to do, the parameters and size of parameters to
 *           switch between both processors after the PAR call. See docs for
 *           more info.
 */
#ifdef CCMPI

#define PAR(f1, v1, s1, f2, v2, s2)							\
{											\
  bit--;										\
  NUMPROCESSORS >>= 1;									\
  if ((LLNAME & (1 << bit)) == 0) {							\
     f1;										\
     MPI_Sendrecv(&s1, 1, MPI_INT, LLPARTNER(bit), LLCTAG, 				\
                  &s2, 1, MPI_INT, LLPARTNER(bit), LLCTAG, LLC_COMM, &status);	        \
     MPI_Sendrecv(v1, s1, MPI_CHAR, LLPARTNER(bit), LLCTAG, 				\
                  v2, s2, MPI_CHAR, LLPARTNER(bit), LLCTAG, LLC_COMM, &status);	        \
  }											\
  else {										\
     f2;										\
     MPI_Sendrecv(&s2, 1, MPI_INT, LLPARTNER(bit), LLCTAG, 				\
                  &s1, 1, MPI_INT, LLPARTNER(bit), LLCTAG, LLC_COMM, &status);	        \
     MPI_Sendrecv(v2, s2, MPI_CHAR, LLPARTNER(bit), LLCTAG, 				\
                  v1, s1, MPI_CHAR, LLPARTNER(bit), LLCTAG, LLC_COMM, &status);	        \
  }                                                                                     \
  bit++;				                                               	\
  NUMPROCESSORS <<= 1;                                                                  \
}

#else /* PVM, TRANSPUTER */

#define PAR(f1, p1, s1, f2, p2, s2) \
{ \
  char *lltemp;\
  int lltempsize, llsize2;\
  \
	NUMPROCESSORS /= 2; \
	bit--; \
	if ((LLNAME & (1 << bit)) == 0) \
	{ \
		f1; \
		RECEIVEINT(bit, lltempsize); \
		if (lltempsize != 0) { \
			if ((lltemp = (char *)malloc(lltempsize)) == NULL) abort(); \
			RECEIVE(bit, lltemp, lltempsize); \
			} \
		s2 = lltempsize / LLSIZE;\
		llsize2 = s1 * LLSIZE; \
		SENDINT(bit, llsize2); \
		if (s1 != 0) \
			SEND(bit, p1, llsize2); \
		if (lltempsize != 0) { \
			memcpy(p2,lltemp,lltempsize);\
			free(lltemp);\
			} \
	} \
	else \
	{ \
		f2; \
		lltempsize = s2 * LLSIZE; \
		SENDINT(bit, lltempsize); \
		if (lltempsize != 0) \
			SEND(bit, p2, lltempsize); \
		RECEIVEINT(bit, llsize2); \
		s1 = llsize2 / LLSIZE; \
		if (llsize2 != 0) \
			RECEIVE(bit, p1, llsize2); \
	} \
	bit++; \
	NUMPROCESSORS *= 2; \
}
#endif  /* ifdef CCMPI */

#define PARG(f1, p1, s1, f2, p2, s2) \
{ \
  int lli, llaux; \
  TYPENODE *llptr_aux; \
	NUMPROCESSORS /= 2; \
	bit--; \
  if ((llptr_aux = (TYPENODE *) malloc (sizeof (TYPENODE))) == NULL) abort(); \
	if ((LLNAME & (1 << bit)) == 0) \
	{ \
		f1; \
	 llptr_aux = p1; \
	 SENDINT(bit, s1); \
	 for (lli = 0; lli < s1; lli++) \
	 { \
		SEND(bit, READ (&llptr_aux), sizeof (TYPENODE)); \
	 } \
	 RECEIVEINT(bit, llaux); \
	 for (lli = 0; lli < llaux; lli++) \
	 { \
		RECEIVE(bit, llptr_aux, sizeof (TYPENODE)); \
		WRITE(&p2, llptr_aux->data); \
	 } \
	 s2 += llaux; \
	} \
	else \
	{ \
		f2; \
	 RECEIVEINT(bit, llaux); \
	 for (lli = 0; lli < llaux; lli++) \
	 { \
		RECEIVE (bit, llptr_aux, sizeof (TYPENODE)); \
		WRITE(&p1, llptr_aux->data); \
	 } \
	 s1 += llaux; \
	 llptr_aux = p2; \
	 SENDINT (bit, s2); \
	 for (lli = 0; lli < s2; lli++) \
	 { \
		SEND (bit, READ (&llptr_aux), sizeof (TYPENODE)); \
	 } \
  } \
  free (llptr_aux); \
	bit++; \
	NUMPROCESSORS *= 2; \
}

#define PARB(arg1, arg2, arg3, arg4, arg5, arg6) \
{ \
  char *lltemp;\
  int lltempsize;\
  \
	NUMPROCESSORS /= 2; \
	bit--; \
	if ((LLNAME & (1 << bit)) == 0) \
	{ \
		arg1; \
		RECEIVEINT(bit, lltempsize); \
		if ((lltemp = (char *)malloc(lltempsize)) == NULL) abort(); \
	 if (lltempsize != 0) \
		RECEIVE(bit, lltemp, lltempsize); \
		SENDINT(bit, arg3); \
	 if (arg3 != 0) \
		SEND(bit, arg2, arg3); \
		arg6 = lltempsize;\
	 if (arg6 != 0) \
		memcpy(arg5,lltemp,lltempsize);\
		free(lltemp);\
	} \
	else \
	{ \
		arg4; \
		SENDINT(bit, arg6); \
    if (arg6 != 0) \
      SEND(bit, arg5, arg6); \
		RECEIVEINT(bit, arg3); \
    if (arg3 != 0) \
  		RECEIVE(bit, arg2, arg3); \
	} \
	bit++; \
	NUMPROCESSORS *= 2; \
}


/* --- PAR call with no data interchange ---
 * RECEIVES: Both function calls, f1 and f2
 * RETURNS: Nothing
 */
#define SPLIT(f1, f2) \
{ \
	NUMPROCESSORS /= 2; \
	bit--; \
	if ((LLNAME & (1 << bit)) == 0) \
	{ \
		f1; \
	} \
	else \
	{ \
		f2; \
	} \
	bit++; \
	NUMPROCESSORS *= 2; \
}

/* --- PAR call with data exchange from LEFT processor to RIGHT processor --- */
#define PARLEFT(f1, p1, s1, f2) \
{ \
	NUMPROCESSORS /= 2; \
	bit--; \
	if ((LLNAME & (1 << bit)) == 0) \
	{ \
		f1; \
		RECEIVE(bit, p1, s1); \
	} \
	else \
	{ \
		f2; \
		SEND(bit, p1, s1); \
	} \
	bit++; \
	NUMPROCESSORS *= 2; \
}

/* --- PAR call with data exchange from RIGHT processor to LEFT processor --- */
#define PARRIGHT(f1, f2, p2, s2) \
{ \
	NUMPROCESSORS /= 2; \
	bit--; \
	if ((LLNAME & (1 << bit)) == 0) \
	{ \
		f1; \
		SEND(bit, p2, s2); \
	} \
	else \
	{ \
		f2; \
		RECEIVE(bit, p2, s2); \
	} \
	bit++; \
	NUMPROCESSORS *= 2; \
}


/* --- PAR Virtual call, not knowing the # of available processors ---
 * NOTES: If there is more than 1 processor, just calls the PAR macro
 *		    else, calls f1 and f2 sequentially.
 */
#define PARVIRTUAL(f1, p1, s1, f2, p2, s2) \
{ \
	if (NUMPROCESSORS > 1) \
  { \
		PAR(f1, p1, s1, f2, p2, s2); \
  } \
	else \
	{ \
		f1; \
		f2; \
	} \
}

/* --- PAR call for 3 processes --- */
#define PAR3(f1, p1, s1, f2, p2, s2, f3, p3, s3) \
{ \
	NUMPROCESSORS /= 2; \
	bit--; \
	if ((LLNAME & (1 << bit)) == 0) \
	{ \
		f1; \
		RECEIVE(bit, p2, s2); \
		RECEIVE(bit, p3, s3); \
		SEND(bit, p1, s1); \
	} \
	else \
	{ \
		NUMPROCESSORS /= 2; \
		bit--; \
		if ((LLNAME & (1 << bit)) == 0) \
		{ \
			f2; \
			RECEIVE(bit, p3, s3); \
			SEND(bit, p2, s2); \
		} \
		else \
		{ \
			f3; \
			SEND(bit, p3, s3); \
			RECEIVE(bit, p2, s2); \
		} \
		bit++; \
		SEND(bit, p2, s2); \
		SEND(bit, p3, s3); \
		RECEIVE(bit, p1, s1); \
		NUMPROCESSORS *= 2; \
	} \
	bit++; \
	NUMPROCESSORS *= 2; \
}

/* --- PAR call for 4 processes --- */
#define PAR4(f1, p1, s1, f2, p2, s2, f3, p3, s3, f4, p4, s4) \
{ \
	NUMPROCESSORS /= 2; \
	bit--; \
	if ((LLNAME & (1 << bit)) == 0) \
	{ \
		NUMPROCESSORS /= 2; \
		bit--; \
		if ((LLNAME & (1 << bit)) == 0) \
		{ \
			f1; \
			RECEIVE(bit, p2, s2); \
			SEND(bit, p1, s1); \
		} \
		else \
		{ \
			f2; \
			SEND(bit, p2, s2); \
			RECEIVE(bit, p1, s1); \
		} \
		bit++; \
		RECEIVE(bit, p3, s3); \
		RECEIVE(bit, p4, s4); \
		SEND(bit, p1, s1); \
		SEND(bit, p2, s2); \
		NUMPROCESSORS *= 2; \
	} \
	else \
	{ \
		NUMPROCESSORS /= 2; \
		bit--; \
		if ((LLNAME & (1 << bit)) == 0) \
		{ \
			f3; \
			RECEIVE(bit, p4, s4); \
			SEND(bit, p3, s3); \
		} \
		else \
		{ \
			f4; \
			SEND(bit, p4, s4); \
			RECEIVE(bit, p3, s3); \
		} \
		bit++; \
		SEND(bit, p3, s3); \
		SEND(bit, p4, s4); \
		RECEIVE(bit, p1, s1); \
		RECEIVE(bit, p2, s2); \
		NUMPROCESSORS *= 2; \
	} \
	bit++; \
	NUMPROCESSORS *= 2; \
}


/* --- Makes some data swapping in the <bit> dimension ---
 * RECEIVES: the <bit> dimension, and both data to swap
 * TRON: Eliminar el buffering en la versin PVM
 */
#define SWAP(bit, ptr1, size1, ptr2, size2) \
{ \
  char *lltemp;\
  int lltempsize;\
  \
	if ((LLNAME & (1 << bit)) == 0) \
	{ \
		RECEIVEINT(bit, lltempsize); \
		if ((lltemp = (char *)malloc(lltempsize)) == NULL) abort(); \
    if (lltempsize != 0) \
		RECEIVE(bit, lltemp, lltempsize); \
		SENDINT(bit, size1); \
    if (size1 != 0) \
		SEND(bit, ptr1, size1); \
		size2 = lltempsize;\
	 if (size2 != 0) \
		memcpy(ptr2,lltemp,lltempsize);\
		free(lltemp);\
	} \
	else \
	{ \
		SENDINT(bit, size2); \
	 if (ptr2 != 0) \
		SEND(bit, ptr2, size2); \
		RECEIVEINT(bit, size1); \
	 if (size1 != 0) \
		RECEIVE(bit, ptr1, size1); \
	} \
}
/******* End of Level 2 Functions (parallel data interchange related) ******/

#endif /* --- If not compiling a sequential model (ifndef CCSEQ)--- */

#endif /* #ifndef LLCINCLUDED */
